# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""SOC Entity sub-package."""

# flake8: noqa: F401
from .incident import Incident
from .sentinel_alert import SentinelAlert
