# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Initialization sub-package.

This is a sub-package that handles initialization of the package.

- nbinit - notebook initialization
- azure_ml_tools - configuration and helpers for AML workspaces

"""
